@@ -11,6 +11,7 @@ import warnings
 from collections import Counter
 from datetime import datetime
 from functools import lru_cache
+from glob import glob
 from itertools import groupby
 from typing import Iterable, Set
 
@@ -46,6 +47,12 @@ class Gbasf2Process(BatchProcess):
           task with the same project name is running, this b2luigi gbasf2 wrapper will recognize that
           and instead of resubmitting a new project, continue monitoring the running project.
 
+          .. hint::
+            The outputs of gbasf2 tasks can be a bit overwhelming, so I recommend using the
+            :ref:`central scheduler <central-scheduler-label>`
+            which provides a nice overview of all tasks in the browser, including a status/progress
+            indicator how many jobs in a gbasf2 project are already done.
+
         - **Automatic download of datasets and logs**
 
           If all jobs had been successful, it automatically downloads the output dataset and
@@ -127,10 +134,15 @@ class Gbasf2Process(BatchProcess):
 
         Other not required, but noteworthy settings are:
 
-        - ``gbasf2_install_directory``: Defaults to ``~/gbasf2KEK``. If you installed gbasf2 into another
-          location, you have to change that setting accordingly.
+        - ``gbasf2_install_directory``: Directory where gbasf2 is installed.
+            Defaults to ``/cvmfs/belle.kek.jp/grid/gbasf2/pro/../..``, which should point to the latest gbasf2 release.
         - ``gbasf2_release``: Defaults to the release of your currently set up basf2 release.
           Set this if you want the jobs to use another release on the grid.
+        - ``gbasf2_proxy_lifetime``: Defaults to 24. When initializing a proxy, set the
+          lifetime to this number of hours.
+        - ``gbasf2_min_proxy_lifetime``: Defaults to 0. During processing, prompt user
+          to reinitialize proxy if remaining proxy lifetime drops below this number of
+          hours.
         - ``gbasf2_print_status_updates``: Defaults to ``True``. By setting it to ``False`` you can turn off the
           printing of of the job summaries, that is the number of jobs in different states in a gbasf2 project.
         - ``gbasf2_max_retries``: Default to 0. Maximum number of times that each job in the project can be automatically
@@ -292,7 +304,15 @@ class Gbasf2Process(BatchProcess):
                 return JobStatus.running
             return JobStatus.aborted
 
+        # task is running
         if n_in_final_state < n_jobs:
+            # try updating progressbar for central scheduler web view
+            percentage = n_done / n_jobs * 100
+            self._scheduler.set_task_progress_percentage(self.task.task_id, percentage)
+
+            status_message = "\n".join(f"{status}: {n_jobs}" for status, n_jobs in n_jobs_by_status.items())
+            self._scheduler.set_task_status_message(self.task.task_id, status_message)
+
             return JobStatus.running
 
         # Require all jobs to be done for project success, any job failure results in a failed project
@@ -359,9 +379,9 @@ class Gbasf2Process(BatchProcess):
 
         if jobs_hitting_max_n_retries:
             warnings.warn(
-               f"Reached maximum number of rescheduling tries ({self.max_retries}) for following jobs:\n\t" +
-               "\n\t".join(str(j) for j in jobs_hitting_max_n_retries) + "\n",
-               RuntimeWarning
+                f"Reached maximum number of rescheduling tries ({self.max_retries}) for following jobs:\n\t" +
+                "\n\t".join(str(j) for j in jobs_hitting_max_n_retries) + "\n",
+                RuntimeWarning
             )
             return False
 
@@ -402,6 +422,8 @@ class Gbasf2Process(BatchProcess):
         # copies it into the same directory in the grid input sandbox as the
         # steering file.
         pickle_file_symlink_destination = os.path.basename(self.pickle_file_path)
+        if os.path.islink(pickle_file_symlink_destination):
+            os.remove(pickle_file_symlink_destination)
         try:
             os.symlink(self.pickle_file_path, pickle_file_symlink_destination, target_is_directory=False)
             run_with_gbasf2(gbasf2_command)
@@ -480,7 +502,7 @@ class Gbasf2Process(BatchProcess):
         # gbasf2 job priority
         priority = get_setting("gbasf2_priority", default=False, task=self.task)
         if priority is not False:
-            if not (0 <= priority <= 10):
+            if not 0 <= priority <= 10:
                 raise ValueError("Priority should be integer between 0 and 10.")
             gbasf2_command_str += f" --priority {priority} "
 
@@ -557,10 +579,10 @@ class Gbasf2Process(BatchProcess):
                 "but gbasf2 batch only supports root outputs"
             )
         dataset_query_string = \
-            f"/belle/user/{self.dirac_user}/{self.gbasf2_project_name}/sub00/{output_file_stem}_*{output_file_ext}"
+            f"/belle/user/{self.dirac_user}/{self.gbasf2_project_name}/sub*/{output_file_stem}_*{output_file_ext}"
         return dataset_query_string
 
-    def _local_gb2_dataset_is_complete(self, output_file_name: str, check_temp_dir: bool = False, verbose: bool = False) -> bool:
+    def _local_gb2_dataset_is_complete(self, output_file_name: str, check_temp_dir: bool = False) -> bool:
         """
         Helper method that returns ``True`` if the download of the gbasf2
         dataset for the output ``output_file_name`` is complete.
@@ -572,44 +594,38 @@ class Gbasf2Process(BatchProcess):
             check_temp_dir: Instead of checking the final output path, check whether the download into the
                 temporary ("partial") directory is complete. This function is usually called with this
                 argument set to ``True``, to check whether the dataset can be moved to its final output path.
-            verbose: If true, print filenames in datasets.
         """
         # first get the local set of files in the dataset for `output_file_name`
         task_output_dict = flatten_to_dict(self.task.output())
         output_target = task_output_dict[output_file_name]
         output_dir_path = output_target.path
         if check_temp_dir:
-            output_dir_path = os.path.join(output_dir_path + ".partial", self.gbasf2_project_name, 'sub00')
-        if not os.path.isdir(output_dir_path):
-            raise FileNotFoundError(errno.ENOTDIR, os.strerror(errno.ENOTDIR), output_dir_path)
-        downloaded_dataset_basenames = os.listdir(output_dir_path)
+            # files in the temporary download directory
+            glob_expression = os.path.join(f"{output_dir_path}.partial", self.gbasf2_project_name, "sub*", "*.root")
+            downloaded_dataset_basenames = [os.path.basename(fpath) for fpath in glob(glob_expression)]
+        else:
+            # file in the final output directory
+            downloaded_dataset_basenames = os.listdir(output_dir_path)
+        if not downloaded_dataset_basenames:
+            return False
 
-        # not get the remote set of grid file names for the gbasf2 project output matching output_file_name
+        # get the remote set of grid file names for the gbasf2 project output matching output_file_name
         ds_query_string = self._get_gbasf2_dataset_query(output_file_name)
-        ds_list_command = shlex.split(f"gb2_ds_list {ds_query_string}")
-        output_dataset_grid_filepaths = run_with_gbasf2(ds_list_command, capture_output=True).stdout.splitlines()
+        output_dataset_grid_filepaths = query_lpns(ds_query_string)
         output_dataset_basenames = [os.path.basename(grid_path) for grid_path in output_dataset_grid_filepaths]
         # remove duplicate LFNs that gb2_ds_list returns for outputs from rescheduled jobs
         output_dataset_basenames = get_unique_lfns(output_dataset_basenames)
         # check if local and remote datasets are equal
         if set(output_dataset_basenames) == set(downloaded_dataset_basenames):
             return True
-        if verbose:
-            print(
-                "\nDownloaded files:\n{}".format("\n".join(downloaded_dataset_basenames)) +
-                "\nFiles on the grid:\n{}".format("\n".join(output_dataset_basenames))
-            )
+        missing_files = list(set(output_dataset_basenames).difference(downloaded_dataset_basenames))
+        superfluous_files = list(set(downloaded_dataset_basenames).difference(output_dataset_basenames))
+        if missing_files:
+            warnings.warn("\nFiles missing in download:\n{}".format("\n".join(missing_files)), RuntimeWarning)
+        if superfluous_files:
+            warnings.warn("\nFiles superfluous in download:\n{}".format("\n".join(superfluous_files)), RuntimeWarning)
         return False
 
-    def _failed_files_from_dataset_download(self, stdout):
-        """
-        Parse stdout from gb2_ds_get dataset download command to extract LFN's of failed file downloads.
-        """
-        failed_files = stdout.split('Failed files:')[-1].\
-            split("Files with duplicated jobID, not downloaded:")[0].strip().split('\n')
-        failed_files = [line for line in failed_files if len(line.strip()) > 0]
-        return failed_files
-
     def _download_dataset(self):
         """
         Download the task outputs from the gbasf2 project dataset.
@@ -638,18 +654,25 @@ class Gbasf2Process(BatchProcess):
             tmp_output_dir_path = f"{output_dir_path}.partial"
             os.makedirs(tmp_output_dir_path, exist_ok=True)
 
-            # Need a file to repeat download for FAILED ones only
+            # Need a set files to repeat download for FAILED ones only
             monitoring_failed_downloads_file = os.path.join(tmp_output_dir_path, "failed_files.txt")
+            monitoring_download_file_stem, monitoring_downloads_file_ext = os.path.splitext(monitoring_failed_downloads_file)
+            old_monitoring_failed_downloads_file = f"{monitoring_download_file_stem}_old{monitoring_downloads_file_ext}"
 
-            # In case of first download, this file does not exist
+            # In case of first download, the file 'monitoring_failed_downloads_file' does not exist
             if not os.path.isfile(monitoring_failed_downloads_file):
 
-                ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string}")
+                ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string} "
+                                             f"--failed_lfns {monitoring_failed_downloads_file}")
                 print("Downloading dataset with command ", " ".join(ds_get_command))
 
             # Any further time is based on the list of files from failed downloads
             else:
-                ds_get_command = shlex.split(f"gb2_ds_get --force --input_dslist {monitoring_failed_downloads_file}")
+                # Rename current 'monitoring_failed_downloads_file' to reuse the path in case the download fails again
+                os.rename(monitoring_failed_downloads_file, old_monitoring_failed_downloads_file)
+                ds_get_command = shlex.split(f"gb2_ds_get --force {dataset_query_string} "
+                                             f"--input_dslist {old_monitoring_failed_downloads_file} "
+                                             f"--failed_lfns {monitoring_failed_downloads_file}")
                 print("Downloading remaining files from dataset with command ", " ".join(ds_get_command))
 
             stdout = run_with_gbasf2(ds_get_command, cwd=tmp_output_dir_path, capture_output=True).stdout
@@ -657,19 +680,22 @@ class Gbasf2Process(BatchProcess):
             if "No file found" in stdout:
                 raise RuntimeError(f"No output data for gbasf2 project {self.gbasf2_project_name} found.")
 
-            failed_files = self._failed_files_from_dataset_download(stdout)
-            if failed_files:
-                with open(monitoring_failed_downloads_file, 'w') as ffs:
-                    ffs.write("\n".join(failed_files))
-            else:
-                try:
-                    os.remove(monitoring_failed_downloads_file)
-                except OSError as e:
-                    if e.errno != errno.ENOENT:  # errno.ENOENT = no such file or directory
-                        raise  # re-raise exception if a different error occurred
+            # Check, whether a file with failed lfns was created
+            failed_files = []
+            if os.path.isfile(monitoring_failed_downloads_file):
+                with open(monitoring_failed_downloads_file, "r") as failed_downloads_fileobj:
+                    failed_files = [f.strip() for f in failed_downloads_fileobj.readlines()]
 
-            tmp_output_dir = os.path.join(tmp_output_dir_path, self.gbasf2_project_name, 'sub00')
-            if not self._local_gb2_dataset_is_complete(output_file_name, check_temp_dir=True, verbose=True):
+            if not failed_files:
+                # In case all downloads were successful:
+                # remove 'old_monitoring_failed_downloads_file' and 'monitoring_failed_downloads_file'
+                if os.path.isfile(old_monitoring_failed_downloads_file):
+                    os.remove(old_monitoring_failed_downloads_file)
+                if os.path.isfile(monitoring_failed_downloads_file):
+                    os.remove(monitoring_failed_downloads_file)
+
+            tmp_output_dir = os.path.join(tmp_output_dir_path, self.gbasf2_project_name)
+            if not self._local_gb2_dataset_is_complete(output_file_name, check_temp_dir=True):
                 raise RuntimeError(
                     f"Download incomplete. The downloaded set of files in {tmp_output_dir} is not equal to the " +
                     f"list of dataset files on the grid for project {self.gbasf2_project_name}.",
@@ -679,7 +705,12 @@ class Gbasf2Process(BatchProcess):
                   f"Moving output files to directory: {output_dir_path}")
             if os.path.exists(output_dir_path):
                 shutil.rmtree(output_dir_path)
-            shutil.move(src=tmp_output_dir, dst=output_dir_path)
+            # sub00 and other sub<xy> directories in case of other datasets
+            sub_directories = glob(os.path.join(tmp_output_dir, "sub*"))
+            for sub_dir in sub_directories:
+                if not os.path.isdir(sub_dir):
+                    raise NotADirectoryError(f"{sub_dir} is not a directory.")
+                shutil.move(src=sub_dir, dst=output_dir_path)
             shutil.rmtree(tmp_output_dir_path)
 
     def _download_logs(self):
@@ -751,26 +782,13 @@ class Gbasf2GridProjectTarget(Target):
 
 def check_dataset_exists_on_grid(gbasf2_project_name, dirac_user=None):
     """
-    Use ``gb2_ds_list`` command to see if an output dataset exists for the gbasf2 project
+    Check if an output dataset exists for the gbasf2 project
     """
-    if dirac_user is None:
-        dirac_user = get_dirac_user()
-    ds_list_command = shlex.split(f"gb2_ds_list --user {dirac_user} {gbasf2_project_name}")
-    output_dataset_str = run_with_gbasf2(ds_list_command, capture_output=True).stdout
-    if "No datasets" in output_dataset_str:
-        return False
-    output_lines_are_paths = all(os.path.abspath(line) for line in output_dataset_str.strip().splitlines())
-    if not output_lines_are_paths:
-        warnings.warn(
-            "The output of ``{' '.join(ds_list_command)}`` contains lines that are not grid paths:\n" +
-            output_dataset_str,
-            category=RuntimeWarning
-        )
-        return False
-    return True
+    lpns = query_lpns(gbasf2_project_name, dirac_user=dirac_user)
+    return len(lpns) > 0
 
 
-def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None):
+def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None, n_retries=5):
     """
     Returns a dictionary for all jobs in the project with a structure like the
     following, which I have taken and adapted from an example output::
@@ -808,8 +826,18 @@ def get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user=None):
             "Probably there was an error during the project submission when running the gbasf2 command.\n"
         )
     job_status_json_string = proc.stdout
-    job_status_dict = json.loads(job_status_json_string)
-    return job_status_dict
+    try:
+        job_status_dict = json.loads(job_status_json_string)
+        return job_status_dict
+    except json.decoder.JSONDecodeError as err:
+        warnings.warn(
+            " Failed to decode job status json string: \n" +
+            job_status_json_string
+        )
+        if n_retries > 0:
+            print(f"Trying to get new job status ({n_retries - 1} retries left)")
+            return get_gbasf2_project_job_status_dict(gbasf2_project_name, dirac_user, n_retries=n_retries - 1)
+        raise err
 
 
 def check_project_exists(gbasf2_project_name, dirac_user=None):
@@ -868,12 +896,18 @@ def get_gbasf2_env(gbasf2_install_directory=None):
     :return: Dictionary containing the  environment that you get from sourcing the gbasf2 setup script.
     """
     if gbasf2_install_directory is None:
-        gbasf2_install_directory = get_setting("gbasf2_install_directory", default="~/gbasf2KEK")
+        # Use the latest gbasf2 release on CVMFS as the default gbasf2 install directory.
+        # To get the directory of the latest release, take the parent of the "pro"
+        # symlink which points to a sub-directory of the latest release.
+        default_gbasf2_install_directory = os.path.realpath(
+            os.path.join("/cvmfs/belle.kek.jp/grid/gbasf2/pro", os.pardir, os.pardir)
+        )
+        gbasf2_install_directory = get_setting("gbasf2_install_directory", default=default_gbasf2_install_directory)
     gbasf2_setup_path = os.path.join(gbasf2_install_directory, "BelleDIRAC/gbasf2/tools/setup")
-    if not os.path.isfile(os.path.expanduser(gbasf2_setup_path)):
+    if not os.path.isfile(gbasf2_setup_path):
         raise FileNotFoundError(
-            f"Could not find gbasf2 setup files in ``{gbasf2_install_directory}``.\n" +
-            "Make sure to that gbasf2 is installed at that location."
+            f"Could not find gbasf2 setup file at:\n{gbasf2_setup_path}.\n"
+            f"Make sure that `gbasf2_install_directory` is set correctly. Current setting:\n{gbasf2_install_directory}.\n"
         )
     # complete bash command to set up the gbasf2 environment
     # piping output to /dev/null, because we want that our final script only prints the ``env`` output
@@ -884,35 +918,78 @@ def get_gbasf2_env(gbasf2_install_directory=None):
         echo_gbasf2_env_command, check=True, stdout=subprocess.PIPE, encoding="utf-8"
     ).stdout
     gbasf2_env = dict(line.split("=", 1) for line in gbasf2_env_string.splitlines())
+    # The gbasf2 setup script on sets HOME to /ext/home/ueda if it's unset,
+    # which later causes problems in the gb2_proxy_init subprocess. Therefore,
+    # reset it to the caller's HOME.
+    gbasf2_env["HOME"] = os.getenv("HOME")
     return gbasf2_env
 
 
+def get_proxy_info():
+    """Run ``gbasf2_proxy_info.py`` to retrieve a dict of the proxy status."""
+    proxy_info_script_path = os.path.join(
+        os.path.dirname(os.path.realpath(__file__)),
+        "gbasf2_utils/gbasf2_proxy_info.py"
+    )
+
+    # Setting ``initalize_proxy=False`` is vital here, otherwise we get an infinite loop
+    try:
+        proc = run_with_gbasf2(
+            [proxy_info_script_path],
+            capture_output=True,
+            ensure_proxy_initialized=False,
+        )
+    except subprocess.CalledProcessError:
+        return {}
+
+    return json.loads(proc.stdout)
+
+
 def get_dirac_user():
     """Get dirac user name"""
-    proxy_info_str = run_with_gbasf2(["gb2_proxy_info"], capture_output=True).stdout
-    for line in proxy_info_str.splitlines():
-        if line.startswith("username"):
-            dirac_user = line.split(":", 1)[1].strip()
-            return dirac_user
-    raise RuntimeError("Could not obtain dirac user name from `gb2_proxy_init` output.")
+    try:
+        return get_proxy_info()["username"]
+    except KeyError as err:
+        raise RuntimeError(
+            "Could not obtain dirac user name from `gb2_proxy_init` output."
+        ) from err
 
 
 def setup_dirac_proxy():
     """
     Runs ``gb2_proxy_init -g belle`` if there's no active dirac proxy. If there is, do nothing.
     """
-    check_proxy_initizalized_script_path = os.path.join(
-        os.path.dirname(os.path.realpath(__file__)),
-        "gbasf2_utils/check_if_dirac_proxy_is_initialized.py"
-    )
     # first run script to check if proxy is already alive or needs to be initalized
-    # setting ``initalize_proxy=False`` is vital here, otherwise we get an infinite loop
-    proc = run_with_gbasf2([check_proxy_initizalized_script_path], ensure_proxy_initialized=False, check=False)
-    # if returncode of the script is 0, that means that proxy is already alive
-    if not proc.returncode:
+    if get_proxy_info().get("secondsLeft", 0) > 3600 * get_setting("gbasf2_min_proxy_lifetime", default=0):
         return
-    # initiallize proxy
-    run_with_gbasf2(shlex.split("gb2_proxy_init -g belle"), ensure_proxy_initialized=False)
+    # initialize proxy
+    lifetime = get_setting("gbasf2_proxy_lifetime", default=24)
+    if not isinstance(lifetime, int) or lifetime <= 0:
+        warnings.warn("Setting 'gbasf2_proxy_lifetime' should be a positive integer.", RuntimeWarning)
+    hours = int(lifetime)
+    run_with_gbasf2(shlex.split(f"gb2_proxy_init -g belle -v {hours}:00"), ensure_proxy_initialized=False)
+
+
+def query_lpns(ds_query, dirac_user=None):
+    """
+    Query DIRAC for LPNs matching query, and return them as a list.
+
+    This function exists to avoid manual string parsing of ``gb2_ds_list``.
+    """
+    if dirac_user is None:
+        dirac_user = get_dirac_user()
+
+    ds_list_script_path = os.path.join(
+        os.path.dirname(os.path.realpath(__file__)), "gbasf2_utils/gbasf2_ds_list.py"
+    )
+
+    proc = run_with_gbasf2(
+        [ds_list_script_path, "--dataset", ds_query, "--user", dirac_user],
+        capture_output=True,
+        ensure_proxy_initialized=True,
+    )
+
+    return json.loads(proc.stdout)
 
 
 def get_unique_project_name(task):
